{% set title = title|default(_fsdomain("Reset password")) %}
{% extends "security/base.html" %}
{% from "security/_macros.html" import render_field_with_errors, render_field, render_field_errors, render_form_errors %}

{% block content %}
  {% include "security/_messages.html" %}
  <h1>{{ _fsdomain('Reset password') }}</h1>
  <form action="{{ url_for_security('reset_password', token=reset_password_token) }}" method="post" name="reset_password_form">
    {{ reset_password_form.hidden_tag() }}
    {{ render_form_errors(reset_password_form) }}
    {{ render_field_with_errors(reset_password_form.password) }}
    {{ render_field_with_errors(reset_password_form.password_confirm) }}
    {{ render_field_errors(reset_password_form.csrf_token) }}
    {{ render_field(reset_password_form.submit) }}
  </form>
  {% include "security/_menu.html" %}
{% endblock content %}
