{# This template receives the following context:

  user - the entire user model object
  security - the Flask-Security configuration
  recovery_link - forgot password link if enabled (reset_link below is more useful)
  reset_link - reset link if enabled
  reset_token - this token is part of reset link - but can be used to
    construct arbitrary URLs for redirecting.
  confirmation_link - confirmation link is user not yet confirmed (and enabled)
  confirmation_token

  This template is used when returning generic responses and don't/can't
  provide detailed errors as part of form validation to avoid email/username
  enumeration.
#}
<div>{{ _fsdomain('Hello %(email)s!', email=user.email) }}</div>
<div>{{ _fsdomain('Someone (you?) tried to register this email - which is already in our system.') }}</div>
{% if user.username %}
  <div>
    {{ _fsdomain('This account also has the following username associated with it: %(username)s.', username=user.username) }}
  </div>
{% endif %}
{% if reset_link %}
  <div>{{ _fsdomain('You can use <a href="%(reset_link)s">this link</a> to reset your password.', reset_link=reset_link)|safe }}</div>
{% endif %}
{% if confirmation_link %}
  <div>{{ _fsdomain('You have not confirmed your email address yet - use <a href="%(confirmation_link)s">this link</a> to do so now.', confirmation_link=confirmation_link)|safe }}</div>
{% endif %}
